import pytest
from functools import wraps


def database(transactional=False, reset_sequences=False, autouse=True):
    from pytest_django.fixtures import _django_db_fixture_helper as helper
    @pytest.fixture(autouse=autouse)
    def db(request, django_db_setup, django_db_blocker):
        if reset_sequences: assert transactional, "Reset sequences only works when transactional is True."
        helper(request, django_db_blocker, transactional=transactional, reset_sequences=reset_sequences)
    return db

def authenticated_rest_client(autouse=True):
    @pytest.fixture(autouse=True)
    def authenticated_rest_client_fixture(rest, user):
        def inject_api_token(f):
            @wraps(f)
            def wrapped(*args, **kwargs):
                kwargs['HTTP_X_API_KEY'] = user._token
                return f(*args, **kwargs)
            return wrapped

        rest.get = inject_api_token(rest.get)
        rest.put = inject_api_token(rest.put)
        rest.post = inject_api_token(rest.post)
        rest.delete = inject_api_token(rest.delete)
        return rest
    return authenticated_rest_client_fixture

@pytest.fixture()
def rest():
    from rested.test import Client
    return Client()



from urllib.parse import urlencode
from django.test import TestCase
from rested import json

class RestTestCase(TestCase):

    def get(self, url, params=None, **extra):
        params = '?' + urlencode(params) if params else ''
        response = self.client.get(url + params, **extra)
        response.data = response.json()
        response.status = response.status_code
        return response

    def put(self, url, data=None, params=None, **extra):
        params = '?' + urlencode(params) if params else ''
        response = self.client.put(url + params, json.dumps(data), content_type="application/json", **extra)
        response.data = response.json()
        response.status = response.status_code
        return response

    def post(self, url, data=None, params=None, **extra):
        params = '?' + urlencode(params) if params else ''
        response = self.client.post(url + params, json.dumps(data), content_type="application/json", **extra)
        response.data = response.json()
        response.status = response.status_code
        return response

    def delete(self, url, params=None, **extra):
        params = '?' + urlencode(params) if params else ''
        response = self.client.delete(url + params, **extra)
        response.data = response.json()
        response.status = response.status_code
        return response



from django.db import models
from django.contrib.auth import get_user_model
from django.utils.crypto import get_random_string
from django.contrib.auth.hashers import make_password, check_password


# user mixin for token authentication
class TokenUserMixin(models.Model):
    token = models.CharField(max_length=128, null=True)

    class Meta:
        abstract = True

    def assign_token(self):
        length = 48
        allowed_chars = 'abcdefghjkmnpqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ23456789'
        if self.id is None: raise RuntimeError('User must have an id assigned before token can be generated.')
        token = str(self.id) + ':' + get_random_string(length, allowed_chars)
        self.set_token(token)
        return token

    def set_token(self, raw_token):
        self.token = make_password(raw_token)
        self._token = raw_token

    def check_token(self, raw_token):
        if not self.token: return False
        def setter(raw_token):
            self.set_token(raw_token)
            self._token = None
            self.save(update_fields=["token"])
        return check_password(raw_token, self.token, setter)

# token authentication middleware
class TokenAuthenticationMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.authenticate(request)
        return self.get_response(request)

    @classmethod
    def authenticate(cls, request):
        creds = cls.get_token(request)
        if not creds or ':' not in creds: return
        id, _ = creds.split(':', 1)
        user = get_user_model().objects.filter(id=id).first()
        if not user.check_token(creds): return
        request.csrf_processing_done = True
        request.user = user

    @staticmethod
    def get_token(request):
        return str(request.META.get('HTTP_X_API_KEY', '')).strip()
